
//
// gomoku.cc
// Roy van Hal en Walter Kosters
// 26 maart 2019
// g++ -Wall -O2 -o gomoku gomoku.cc
//

#include <iostream>
#include <cstdlib>
#include <ctime>
using namespace std;

class Basisspeler;

class gomoku {
  public: // everything public, who cares :(
    static const int MAXSIZE = 50;
    static const int BUSY = 0;
    static const int BLACK = 1;
    static const int WHITE = 2;
    static const int DRAW = 3;
    static const int INAROW = 5;
    static const int RESOURCES = 100000;
    char board[MAXSIZE][MAXSIZE]; // the game board
      // board[0][0] is topleft; board[i][j]: row i, column j
      // empty: '.', White: 'W', Black: 'B'
    int height; // actual height of the board
    int width;  // actual width of the board
    int empties; // number of empty positions
    bool playertomove; // true: Black, false: White
    int lasti, lastj; // last move
    Basisspeler* spelers[2]; // the two AI players

    gomoku ( ); // default constructor
    gomoku (int h, int w); // other constructor with height and width
    void cleanboard ( ); // clean the board
    void print ( ); // print the board
    int gamestatus ( ); // status of the game: BUSY, BLACK, WHITE or DRAW
    bool finished ( ); // is game finished?
    int numberofmoves ( ); // how many empty positions?
    int name (char color); // 'W' --> WHITE, 'B' --> BLACK
    int whois ( ); // who is to move?
    void dothemove (int r); // do r-the move, assuming it exists; r >= 1
    void randommove ( ); // do random move, assuming there is at least one
    void firstmove ( ); // do first move, assuming there is at least one
    void printoutcome ( ); // who won?
    int getwinner ( ); // who won?
    void initplayers ( ); // initialize players
};//gomoku

class Basisspeler {
  protected:
    gomoku* spel;
  public:
    virtual int volgendeZet ( ) = 0;
};//Basisspeler

// default constructor
gomoku::gomoku ( ) {
  height = width = 8;
  cleanboard ( );
}//gomoku::gomoku

// other constructor with height and width
gomoku::gomoku (int h, int w) {
  height = h;
  width = w;
  cleanboard ( );
}//gomoku::gomoku

// clean the board
void gomoku::cleanboard ( ) {
  int i, j;
  empties = height * width;
  for ( i = 0; i < height; i++ )
    for ( j = 0; j < width; j++ )
      board[i][j] = '.';
  playertomove = true; // Black begins
  lasti = lastj = -1;
  initplayers ( );
}//gomoku::cleanboard

// print the board
void gomoku::print ( ) {
  int i, j;
  for ( i = 0; i < height; i++ ) {
    for ( j = 0; j < width; j++ )
      cout << " " << board[i][j];
    cout << endl;
  }//for
  if ( playertomove )
    cout << "Black to move" << endl << endl;
  else
    cout << "White to move" << endl << endl;
}//gomoku::print

// 'W' --> WHITE, 'B' --> BLACK
int gomoku::name (char color) {
  if ( color == 'W' )
    return WHITE;
  return BLACK;
}//gomoku::name

// who is to move?
int gomoku::whois ( ) {
  if ( playertomove )
    return BLACK;
  return WHITE;
}//gomoku::whois

// status of the game: BUSY, BLACK, WHITE or DRAW
int gomoku::gamestatus ( ) {
  int i0, j0, counter;
  if( lasti == -1 || lastj == -1 )
    return BUSY;
  char color = board[lasti][lastj];

  i0 = lasti; j0 = lastj + 1; 
  counter = 1;
  while ( j0 < width && board[i0][j0] == color ) {
    j0++;
    counter++;
  }//while
  i0 = lasti; j0 = lastj - 1; 
  while ( j0 >= 0 && board[i0][j0] == color ) {
    j0--;
    counter++;
  }//while
  if ( counter >= INAROW )
    return name (color);

  i0 = lasti + 1; j0 = lastj; 
  counter = 1;
  while ( i0 < height && board[i0][j0] == color ) {
    i0++;
    counter++;
  }//while
  i0 = lasti - 1; j0 = lastj; 
  while ( i0 >= 0 && board[i0][j0] == color ) {
    i0--;
    counter++;
  }//while
  if ( counter >= INAROW )
    return name (color);

  i0 = lasti + 1; j0 = lastj + 1; 
  counter = 1;
  while ( i0 < height && j0 < width && board[i0][j0] == color ) {
    i0++; j0++;
    counter++;
  }//while
  i0 = lasti - 1; j0 = lastj - 1; 
  while ( i0 >= 0 && j0 >= 0 && board[i0][j0] == color ) {
    i0--; j0--;
    counter++;
  }//while
  if ( counter >= INAROW )
    return name (color);

  i0 = lasti + 1; j0 = lastj - 1; 
  counter = 1;
  while ( i0 < height && j0 >= 0 && board[i0][j0] == color ) {
    i0++; j0--;
    counter++;
  }//while
  i0 = lasti - 1; j0 = lastj + 1; 
  while ( i0 >= 0 && j0 < width && board[i0][j0] == color ) {
    i0--; j0++;
    counter++;
  }//while
  if ( counter >= INAROW )
    return name (color);

  if ( empties > 0 )
    return BUSY;
  return DRAW;
}//gomoku::gamestatus

// how many empty positions?
int gomoku::numberofmoves ( ) {
  int i, j, counter = 0;
  for ( i = 0; i < height; i++ )
    for ( j = 0; j < width; j++ )
      if ( board[i][j] == '.' )
        counter++;
  return counter;
}//gomoku::numberofmoves

// do the r-th move, assuming it exists; r >= 1
void gomoku::dothemove (int r) {
  int i, j;
  for ( i = 0; i < height; i++ )
    for ( j = 0; j < width; j++ )
      if ( board[i][j] == '.' ) {
        r--;
	if ( r == 0 ) {
	  if ( playertomove )
	    board[i][j] = 'B';
	  else
	    board[i][j] = 'W';
	  lasti = i;
	  lastj = j;
	  playertomove = ! playertomove;
	  empties--;
	}//if
      }//if
}//gomoku:dothemove

// do random move, assuming there is at least one
void gomoku::randommove ( ) {
  dothemove (rand ( ) % numberofmoves ( ) + 1);
}//gomoku::randommove

// do first move, assuming there is at least one
void gomoku::firstmove ( ) {
  dothemove (1);
}//gomoku::firstmove

// who won?
void gomoku::printoutcome ( ) {
  int gs = gamestatus ( );
  if ( gs == BLACK )
    cout << "Black wins!" << endl;
  else if ( gs == WHITE )
    cout << "White wins!" << endl;
  else
    cout << "Draw!" << endl;
}//gomoku::printoutcome

// who won?
int gomoku::getwinner ( ) {
  int gs = gamestatus ( );
  if ( gs == BLACK )
    return 0;
  if ( gs == WHITE )
    return 1;
  return 2;
}//gomoku::getwinner

// is game finished?
bool gomoku::finished ( ) {
  return (gamestatus ( ) != BUSY);
}//gomoku::finished

//TODO
#include "gomokudommespeler.cc"

#include "gomokuMCspeler.cc"

// initialize players: should comply with these two include files
void gomoku::initplayers ( ) {
  spelers[0] = new Dommespeler (this);
  spelers[1] = new MCspeler (this);
}//gomoku::initplayers
//TODO

int main ( ) {
  //TODO
  int aantalspellen = 5;
  int seed = 17;
  int hoogte = 10;
  int breedte = 10;
  //TODO
  int wins[3] = {0};
  int p;
    
  cout << "Er worden " << aantalspellen << " spelletjes gespeeld." << endl;
  cout << "Bordgrootte is hoogte = " << hoogte 
       << " bij breedte = " << breedte << "." << endl;
  cout << "De random-seed is " << seed << "." << endl;
  srand (seed); // gebruik eventueel time (NULL)

  gomoku B(hoogte,breedte);
  for ( p = 0; p < aantalspellen; p++ ) {
    B.cleanboard ( );
    do {
      if ( B.playertomove )
        B.dothemove (B.spelers[0]->volgendeZet ( ));
      else
        B.dothemove (B.spelers[1]->volgendeZet ( ));
    } while ( ! B.finished ( ) );
    wins[B.getwinner ( )]++;
  }//for

  cout << "Speler 0 wint: " << wins[0] 
       << ", en Speler 1 wint: " << wins[1]
       << ", en " << wins[2] << " remises." << endl;
  return 0;
}//main

